/* Copyright (C) 2015, Armin Biere, JKU Linz. */

static const char * usage =
"usage: ratz [ <option> ... ] [ <input> [ <output> ] ]\n"
"\n"
"where <option> is one of the following:\n"
"\n"
"-h | --help        print this command line option summary\n"
"-v                 enable printing of verbose messages\n"
"\n"
"output format options:\n"
"\n"
"-a | --ascii       original ASCII format\n"
"-b | --binary      fixed 32-bit word binary format\n"
"\n"
"--variable         variable-byte format (default)\n"
"--variable-byte\n"
"\n"
"--sort             keep first literal sort rest increasing by index\n"
"--diff             use difference after 2nd literal after sorting\n"
"                   (default for '-m', implies '-s', not for '-a')\n"
"\n"
"-r | --release     release deleted clauses by reference (hashing)\n"
"-o | --order       order and reencode literals by number occurrences\n"
"\n"
"--no-sort          do not sort (default for '-a' and '-b')\n"
"--no-diff          do not use differences\n"
"\n"
"--big-endian       instead of default little-endian\n"
"                   (can only be used for binary format)\n"
"\n"
"The <input> is supposed to be a DRAT proof file in ASCII format and is\n"
"written in the specified format to the output file.  If <output> is not\n"
"specified then standard output is used and similarly if <input> is not\n"
"specified then standard input is used.  If a file is specified and\n"
"has a name with '.gz', '.bz2', or '.7z' suffix, then input and output\n"
"are decompressed and compressed using helper programs through a pipe.\n"
"The default is the variable-byte output format, unless no output file\n"
"is given and standard output is connected to a terminal, in which case\n"
"the ASCII format is used.\n" 
;

/*------------------------------------------------------------------------*/

#include "typez.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/

#define BINARY_POS_BITS (32 - RATZ_TYPE_BITS)

/*------------------------------------------------------------------------*/

typedef struct Cls {
  unsigned id, hash;
  int size, * lits;
  struct Cls * chain, * prev, * next;
} Cls;

/*------------------------------------------------------------------------*/

static char * input_name, * output_name;
static FILE * input_file, * output_file;
static int close_input, close_output;

static int verbose_level;
static int ascii_format, binary_format, variable_format;
static int sort_format, no_sort_format;
static int diff_format, no_diff_format;
static int big_endian_mode, release_mode, order_mode;

static int * lits, nlits, szlits;
static int saved_pivot_pos = -1;

static int lineno = 1;

static long read_bytes, written_bytes;
static long added_clauses, deleted_clauses, released_clauses;
static long added_lits, deleted_lits, released_lits;

static double start_time;
static long allocated, maxbytes;

static long lookups, collisions;

static Cls ** table, * first, * last;
static unsigned id, ncls, maxcls, sztable;

static int *order, *map, *occs, norder, szorder;

/*------------------------------------------------------------------------*/

static void die (const char * msg, ...) {
  va_list ap;
  fputs ("*** ratz error: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static void wrn (const char * msg, ...) {
  va_list ap;
  fputs ("*** ratz: warning: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
}

static void msg (const char * msg, ...) {
  va_list ap;
  if (!verbose_level) return;
  fputs ("[ratz] ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
}

static void perr (const char * msg, ...) {
  va_list ap;
  fprintf (stderr,
    "*** ratz: parse error in line %d in '%s': ",
    lineno, input_name);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

/*------------------------------------------------------------------------*/

static void incmem (size_t bytes) {
  allocated += bytes;
  if (maxbytes < allocated) maxbytes = allocated;
}

static void decmem (size_t bytes) {
  assert (allocated >= bytes);
  allocated -= bytes;
}

static void * new (size_t bytes, const char * what) {
  void * res = malloc (bytes);
  if (bytes && !res) die ("failed to allocate %s", what);
  incmem (bytes);
  return res;
}

static void del (void * ptr, size_t bytes) {
  decmem (bytes);
  free (ptr);
}

static void * rsz (void * ptr, 
                   size_t oldbytes, size_t newbytes,
		   const char * what) {
  void * res;
  decmem (oldbytes);
  res = realloc (ptr, newbytes);
  if (newbytes && !res) die ("failed to reallocate %s", what);
  incmem (newbytes);
  return res;
}

/*------------------------------------------------------------------------*/

static void push_lit (int lit) {
  if (nlits >= szlits) {
    size_t oldbytes, newbytes;
    if (szlits == (1<<30)) /* see also 'write_binary_type_and_pivot_pos' */
      die ("clause too large (> 2^30 literals)");
    oldbytes = szlits * sizeof *lits;
    szlits = szlits ? 2*szlits : 1;
    newbytes = szlits * sizeof *lits;
    lits = rsz (lits, oldbytes, newbytes, "literal stack");
  }
  lits[nlits++] = lit;
}

/*------------------------------------------------------------------------*/

static void enlarge_table () {
  unsigned i, h, oldsztable = sztable;
  Cls ** oldtable = table, * p, * n;
  size_t bytes;
  sztable = oldsztable ? 2*oldsztable : 1;
  bytes = sztable * sizeof *table;
  table = new (bytes, "enlarged hash table");
  memset (table, 0, bytes);
  for (i = 0; i < oldsztable; i++) {
    for (p = oldtable[i]; p; p = n) {
      n = p->chain;
      h = p->hash & (sztable - 1);
      p->chain = table[h];
      table[h] = p;
    }
  }
  del (oldtable, oldsztable * sizeof *oldtable);
}

static unsigned primes[] = {
  2377883, 2910671, 8458139, 1522097, 5855383, 1934761, 8106251, 1735313,
  4849843, 1518677, 3669583, 4917391, 9100951, 2652569, 8931877, 2203519,
  6318001, 5710781, 3321583, 1044559, 5029309, 5673947, 8541679, 401411,
  1080229, 6393983, 131111, 7201847, 1018177, 4829453, 2543573, 6766979,
  8967853, 7598959, 7202927, 9076241, 9286129, 3776639, 8997587, 7788821,
  3242549, 1942319, 9955541, 1612991, 7080383, 5015201, 9363019, 7163251,
  1051913, 2092049, 5594009, 7054361, 782339, 5105299, 1961983, 2742119,
  6386027, 4482091, 398729, 4672523, 777571, 381011, 1600451, 3978329,
  6888989, 5347123, 6869237, 2724583, 4659367, 2765579, 4500871, 9350699,
};

#define nprimes (sizeof primes / sizeof *primes)

static unsigned hash_lits () {
  unsigned res = 0;
  int i, j = 0;
  for (i = 0; i < nlits; i++) {
    res += primes[j++] * (unsigned) lits[i];
    if (j == nprimes) j = 0;
  }
  return res;
}

static int lits_match_clause (Cls * c, unsigned hash) {
  int i;
  if (c->hash != hash) return 0;
  if (c->size != nlits) return 0;
  for (i = 0; i < nlits; i++)
    if (c->lits[i] != lits[i]) return 0;
  return 1;
}

static Cls ** find_clause (unsigned h, int match) {
  Cls * c, ** p;
  assert (sztable > 0);
  lookups++;
  for (p = &table[h & (sztable-1)];
       (c = *p) && !(match && lits_match_clause (c, h));
       p = &c->chain)
    collisions++;
  return p;
}

static void add_clause () {
  unsigned h = hash_lits ();
  Cls * c, ** p;
  assert (release_mode);
  if (ncls >= sztable) enlarge_table ();
  p = find_clause (h, 0);
  assert (!*p);
  c = new (sizeof *c, "clause");
  *p = c;
  c->hash = h;
  c->size = nlits;
  c->lits = new (nlits * sizeof *c->lits, "clause literals");
  memcpy (c->lits, lits, nlits * sizeof *c->lits);
  c->chain = 0;
  if (++ncls > maxcls) maxcls = ncls;
  c->id = id++;
  if (last) last->next = c;
  else first = c;
  c->prev = last;
  c->next = 0;
  last = c;
}

/*------------------------------------------------------------------------*/

static int read_char () {
  int res = getc_unlocked (input_file);
  if (res == '\n') lineno++;
  if (res != EOF) read_bytes++;
  return res;
}

static void write_char (unsigned char ch) {
  if (putc_unlocked ((int) ch, output_file) == EOF)
    die ("failed to write byte %ld", written_bytes + 1);
  written_bytes++;
}

/*------------------------------------------------------------------------*/

static void init_order () {
  size_t bytes = (szorder = 2) * sizeof (*order);;
  assert (sizeof *order == sizeof *occs);
  assert (sizeof *order == sizeof *map);
  order = new (bytes, "order");
  occs = new (bytes, "occurrences");
  map = new (bytes, "map");
  order[0] = occs[0] = map[0] = 0;
  norder++;
}

static void enlarge_order (int idx) {
  size_t oldbytes, newbytes;
  int newsz;
  assert (0 < idx);
  for (newsz = 2*szorder; newsz <= idx; newsz *= 2)
    assert (newsz < (1<<30));
  oldbytes = szorder * sizeof (*order);
  newbytes = newsz * sizeof (*order);
  szorder = newsz;
  order = rsz (order, oldbytes, newbytes, "order");
  occs = rsz (occs, oldbytes, newbytes, "occurrences");
  map = rsz (map, oldbytes, newbytes, "map");
}

static void new_idx (int idx) {
  while (norder <= idx) {
    occs[norder] = 0;
    order[norder] = norder;
    map[norder] = norder;
    norder++;
  }
}

static void order_lits () {
  int i;
  assert (order_mode);
  if (!norder) init_order ();
  for (i = 0; i < nlits; i++) {
    int idx = abs (lits[i]);
    if (idx >= szorder) enlarge_order (idx);
    if (idx >= norder) new_idx (idx);
    occs[idx]++;
  }
}

static int reencode_lit (int lit) {
  int idx = abs (lit), res;
  assert (0 < idx), assert (idx < norder);
  res = map [idx];
  assert (0 < res), assert (res < norder);
  if (lit < 0) res = -res;
  return res;
}

static void reencode_lits () {
  int i;
  assert (order_mode);
  for (i = 0; i < nlits; i++)
    lits[i] = reencode_lit (lits[i]);
}

/*------------------------------------------------------------------------*/

static int cmp_lit (int a, int b) {
  int i, j, res;
  assert (a), assert (a != INT_MIN);
  assert (b), assert (b != INT_MIN);
  i = abs (a), j = abs (b);
  if ((res = i - j)) return res;	/* smaller index first */
  return (b > 0) - (a > 0);		/* positive literal first */
}

static void remove_duplicate_lits () {
  int i, j = 1, prev;
  assert (sort_format);
  if (!nlits) return;
  prev = lits[0];
  for (i = 1; i < nlits; i++) {
    int next = lits[i];
    assert (cmp_lit (prev, next) <= 0);
    if (prev == next) continue;
    prev = lits[j++] = next;
  }
  nlits = j;
}

static int cmp_lit_qsort (const void * p, const void * q) {
  return cmp_lit (*(int*)p, *(int*)q);
}

static int sort_lits_and_remove_duplicates () {
  int res;
  assert (nlits);
  if (order_mode) order_lits (), reencode_lits ();
  res = lits[0];
  qsort (lits, nlits, sizeof *lits, cmp_lit_qsort);
  remove_duplicate_lits ();
  return res;
}

/*------------------------------------------------------------------------*/

static void write_str (const char * str) {
  const char *p;
  for (p = str; *p; p++)
    write_char (*p);
}

static void ascii_encode_lit (int next) {
  char buf[20];
  assert (next);
  assert (ascii_format);
  sprintf (buf, "%d", next);
  write_str (buf);
}

static void ascii_encode_unsorted_lits (int type) {
  int i;
  assert (ascii_format), assert (!sort_format), assert (!diff_format);
  if (type == RATZ_DEL_TYPE) write_str ("d ");
  else assert (type == RATZ_ADD_TYPE);
  for (i = 0; i < nlits; i++)
    ascii_encode_lit (lits[i]), write_char (' ');
  write_str ("0\n");
}

static void ascii_encode_sorted_lits (int type) {
  assert (ascii_format), assert (sort_format), assert (!diff_format);
  if (type == RATZ_DEL_TYPE) write_str ("d ");
  else assert (type == RATZ_ADD_TYPE);
  if (nlits) {
    int next, i, pivot = sort_lits_and_remove_duplicates ();
    ascii_encode_lit (pivot);
    write_char (' ');
    for (i = 0; i < nlits; i++)
      if ((next = lits[i]) != pivot)
	ascii_encode_lit (next),
	write_char (' ');
  }
  write_str ("0\n");
}

static void ascii_encode_lits (int type) {
  assert (!diff_format);
  if (sort_format) ascii_encode_sorted_lits (type);
  else ascii_encode_unsorted_lits (type);
}

/*------------------------------------------------------------------------*/

static unsigned unsigned_lit (int lit) {
  unsigned res;
  assert (lit), assert (lit != INT_MIN);
  res = abs (lit);
  assert (res <= UINT_MAX/2);
  res *= 2;
  assert (res < UINT_MAX);
  if (lit < 0) res++;
  return res;
}

static unsigned unsigned_difference (int prev, int next) {
  unsigned res;
  assert (sort_format), assert (diff_format);
  assert (next), assert (next != INT_MIN);
  assert (prev != next), assert (prev != INT_MIN);
  res = abs (next);
  assert (res >= abs (prev));
  res -= abs (prev);
  assert (res <= UINT_MAX/2);
  res *= 2;
  assert (res < UINT_MAX);
  if (next < 0) res++;
  return res;
}

static unsigned unsigned_type_and_pivot_pos (int type, int pos) {
  unsigned res;
  assert (type == RATZ_ADD_TYPE ||
          type == RATZ_DEL_TYPE ||
	  type == RATZ_REL_TYPE);
  assert (pos >= 0);
  assert (BINARY_POS_BITS + RATZ_TYPE_BITS == 8*sizeof (unsigned));
  if (pos >= (1 << BINARY_POS_BITS))
    die ("pivot position %d too large in long clause", pos);
  res = type;
  assert (res < (1<<RATZ_TYPE_BITS));
  res |= ((unsigned) pos) << RATZ_TYPE_BITS;
  return res;
}

/*------------------------------------------------------------------------*/

static void write_unsigned (unsigned ulit) {
  assert (binary_format);
  if (big_endian_mode)
    write_char ((ulit >> 24) & 0xff),
    write_char ((ulit >> 16) & 0xff),
    write_char ((ulit >>  8) & 0xff),
    write_char ((ulit >>  0) & 0xff);
  else
    write_char ((ulit >>  0) & 0xff),
    write_char ((ulit >>  8) & 0xff),
    write_char ((ulit >> 16) & 0xff),
    write_char ((ulit >> 24) & 0xff);
}

static void binary_encode_lit (int lit) {
  write_unsigned (unsigned_lit (lit));
}

static void binary_encode_difference (int prev, int next) {
  write_unsigned (unsigned_difference (prev, next));
}

static void binary_encode_type_and_pivot_pos (int type, int pos) {
  write_unsigned (unsigned_type_and_pivot_pos (type, pos));
}

static void binary_encode_unsorted_lits (int type) {
  int i;
  assert (binary_format), assert (!sort_format), assert (!diff_format);
  binary_encode_type_and_pivot_pos (type, 0);
  for (i = 0; i < nlits; i++)
    binary_encode_lit (lits[i]);
  write_unsigned (0);
}

static int pivot_pos () {
  int pivot, res = 0;
  assert (sort_format);
  if (!nlits) return 0;
  if (saved_pivot_pos >= 0) return saved_pivot_pos;
  pivot = sort_lits_and_remove_duplicates ();
  while (lits[res] != pivot)
    res++, assert (res < nlits);
  saved_pivot_pos = res;
  return res;
}

static void binary_encode_sorted_lits (int type) {
  int i;
  assert (binary_format), assert (sort_format), assert (!diff_format);
  binary_encode_type_and_pivot_pos (type, pivot_pos ());
  for (i = 0; i < nlits; i++)
    binary_encode_lit (lits[i]);
  write_unsigned (0);
}

static void binary_encode_differences (int type) {
  int i, prev = 0, next;
  assert (binary_format), assert (sort_format), assert (diff_format);
  binary_encode_type_and_pivot_pos (type, pivot_pos ());
  for (i = 0; i < nlits; i++, prev = next)
    binary_encode_difference (prev, next = lits[i]);
  write_unsigned (0);
}

static void binary_encode_lits (int type) {
  if (diff_format) binary_encode_differences (type);
  else if (sort_format) binary_encode_sorted_lits (type);
  else binary_encode_unsorted_lits (type);
}

/*------------------------------------------------------------------------*/

static void variable_encode_unsigned (unsigned w) {
  while (w > 127)
    write_char (128 | (w & 127)), w >>= 7;
  write_char (w);
}

static void variable_encode_lit (int lit) {
  variable_encode_unsigned (unsigned_lit (lit));
}

static void variable_encode_difference (int prev, int next) {
  variable_encode_unsigned (unsigned_difference (prev, next));
}

static void variable_encode_type_and_pivot_pos (int type, int pos) {
  variable_encode_unsigned (unsigned_type_and_pivot_pos (type, pos));
}

static void variable_encode_unsorted_lits (int type) {
  int i;
  assert (variable_format), assert (!sort_format), assert (!diff_format);
  variable_encode_type_and_pivot_pos (type, 0);
  for (i = 0; i < nlits; i++)
    variable_encode_lit (lits[i]);
  write_char (0);
}

static void variable_encode_sorted_lits (int type) {
  int i;
  assert (variable_format), assert (sort_format), assert (!diff_format);
  variable_encode_type_and_pivot_pos (type, pivot_pos ());
  for (i = 0; i < nlits; i++)
    variable_encode_lit (lits[i]);
  write_char (0);
}

static void variable_encode_differences (int type) {
  int i, prev = 0, next;
  assert (variable_format), assert (sort_format), assert (diff_format);
  variable_encode_type_and_pivot_pos (type, pivot_pos ());
  for (i = 0; i < nlits; i++, prev = next)
    variable_encode_difference (prev, next = lits[i]);
  write_char (0);
}

static void variable_encode_lits (int type) {
  if (diff_format) variable_encode_differences (type);
  else if (sort_format) variable_encode_sorted_lits (type);
  else variable_encode_unsorted_lits (type);
}

/*------------------------------------------------------------------------*/

static void binary_encode_release (int pos, unsigned did) {
  assert (binary_format);
  binary_encode_type_and_pivot_pos (RATZ_REL_TYPE, pos),
  write_unsigned (did);
}

static void variable_encode_release (int pos, unsigned did) {
  assert (variable_format),
  variable_encode_type_and_pivot_pos (RATZ_REL_TYPE, pos),
  variable_encode_unsigned (did);
}

static void encode_release (int pos, unsigned did) {
  assert (release_mode);
  if (binary_format) binary_encode_release (pos, did);
  else assert (variable_format), variable_encode_release (pos, did);
}

/*------------------------------------------------------------------------*/

static int release_clause () {
  unsigned did;
  int pos;
  Cls ** p, * c;
  if (!release_mode) return 0;
  if (!sztable) return 0;
  pos = sort_format ? pivot_pos () : 0;
  p = find_clause (hash_lits (), 1);
  if (!(c = *p)) return 0;
  assert (id >= c->id);
  did = id - c->id;
  encode_release (pos, did);
  assert (ncls > 0);
  *p = c->chain;
  if (c->prev) c->prev->next = c->next;
  else assert (first == c), first = c->next;
  if (c->next) c->next->prev = c->prev;
  else assert (last == c), last = c->prev;
  del (c->lits, c->size * sizeof *c->lits);
  del (c, sizeof *c);
  ncls--;
  return 1;
}

/*------------------------------------------------------------------------*/

static void write_header () {
  unsigned char header;
  if (variable_format) {
    assert (!big_endian_mode);
    if (!diff_format || release_mode) {
			header  = 0;
      if (!diff_format) header |= RATZ_NO_DIFFERENCES_HEADER_BIT;
      if (release_mode) header |= RATZ_RELEASE_HEADER_BIT;
      write_char (RATZ_MAKE_HEADER (header));
    }
  } else if (binary_format) {
	                 header  = RATZ_BINARY_HEADER_BIT;
    if (!diff_format)    header |= RATZ_NO_DIFFERENCES_HEADER_BIT;
    if (big_endian_mode) header |= RATZ_BIG_ENDIAN_HEADER_BIT;
    if (release_mode)    header |= RATZ_RELEASE_HEADER_BIT;
    write_char (RATZ_MAKE_HEADER (header));
  } else {
    assert (ascii_format);
    assert (!big_endian_mode);
    assert (!diff_format);
    assert (!release_mode);
  }
}

static void encode_lits (int type) {
  if (variable_format) variable_encode_lits (type);
  else if (binary_format) binary_encode_lits (type);
  else assert (ascii_format), ascii_encode_lits (type);
}

static void transcode () {
  int type = RATZ_ADD_TYPE;
  write_header ();
  for (;;) {
    int ch = read_char (), lit, sign;
    if (ch == ' ' || ch == '\n') continue;
    if (isdigit (ch)) sign = 1;
    else if (ch == '-') {
      ch = read_char ();
      if (ch == EOF) perr ("unexpected end-of-file after '-'");
      if (!isdigit (ch)) perr ("expected digit after '-'");
      sign = -1;
    } else sign = 0;
    if (sign) {
      lit = ch - '0';
      while (isdigit (ch = read_char ())) {
	int digit = ch - '0';
	if (INT_MAX / 10 < lit) perr ("literal has too many digits");
	lit *= 10;
	if (INT_MAX - digit < lit) perr ("literal too large");
	lit += digit;
      }
      lit *= sign;
      if (ch != ' ' && ch != '\n')
	perr ("expected space or new line after literal %d", lit);
      if (lit) {
	push_lit (lit);
      } else {
	if (type == RATZ_DEL_TYPE && release_clause ())
	  released_clauses++, released_lits += nlits;
	else {
	  encode_lits (type);
	  if (type == RATZ_DEL_TYPE)
	    deleted_clauses++, deleted_lits += nlits;
	  else {
	    assert (type == RATZ_ADD_TYPE);
	    added_clauses++, added_lits += nlits;
	    if (release_mode) add_clause ();
	  }
	} 
	type = RATZ_ADD_TYPE;
	saved_pivot_pos = -1;
	nlits = 0;
      }
      continue;
    }
    if (ch == 'd') {
      if (type == RATZ_DEL_TYPE) perr ("double 'd' without '0'");
      if (nlits) perr ("unexpected 'd' after %d literals", nlits);
      type = RATZ_DEL_TYPE;
      if ((ch = read_char ()) == EOF)
	perr ("unexpected end-of-file after 'd'");
      if (ch != ' ')
	perr ("expected space after 'd'");
      continue;
    }
    if (ch == 'c' || ch == 'p' || ch == 's' || ch == 'v') {
      int tmp;
      while ((tmp = read_char ()) != '\n')
	if (tmp == EOF)
	  perr ("unexpected end-of-file after '%c'", ch);
      continue;
    }
    if (ch == EOF) {
      if (nlits) perr ("unexpected end-of-file (no '0'?)");
      if (type == RATZ_DEL_TYPE) perr ("empty 'd' line without '0'");
      return;
    }
    if (isprint (ch)) perr ("unexpected character '%c'", ch);
    else perr ("unexpected non-printable character (code %d)", ch);
  }
}

/*------------------------------------------------------------------------*/

static long file_bytes (const char * name) {
  struct stat buf;
  if (stat (name, &buf)) return -1;
  return (long) buf.st_size;
}

static double current_time (void) {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0)) res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static double wall_clock_time () { 
  return current_time () - start_time; 
}

static double process_time (void) {
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  if (getrusage (RUSAGE_CHILDREN, &u)) return 0;
  res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

static void stats () {
  long actual_input_bytes, actual_output_bytes;
  long sum = deleted_clauses + released_clauses;
  msg ("added %ld clauses, deleted %ld clauses", added_clauses, sum);
  if (release_mode) {
    msg ("released %ld clauses, %.0f%% of added, %.0f%% of deleted",
    released_clauses,
    added_clauses ? 100.0 * released_clauses/(double)sum : 0,
    sum ? 100.0 * released_clauses/(double)sum : 0);
    msg ("maximum number %d of active clauses %.0f%%",
      maxcls, added_clauses ? 100.0 * maxcls / (double) added_clauses : 0);
  }
  if (release_mode)
    msg ("looked up %ld clauses, %ld collisions %1.2f",
      lookups, collisions, lookups ? collisions / (double) lookups : 0);
  msg ("added %ld literals, %.1f per added clause",
    added_lits, added_clauses ? added_lits / (double) added_clauses : 0);
  msg ("deleted %ld literals, %.1f per deleted clause",
    deleted_lits,
    deleted_clauses ? deleted_lits / (double) deleted_clauses : 0);
  if (release_mode)
    msg ("released %ld literals, %.1f per released clause",
      released_lits,
      released_clauses ? released_lits / (double) released_clauses : 0);
  msg ("read %ld bytes, written %ld bytes",
    read_bytes, written_bytes);
  if (read_bytes)
    msg ("written/read bytes ratio %.2f%%",
      (100.0 * written_bytes) / (double) read_bytes);
  if (input_name) {
    assert (close_input);
    if ((actual_input_bytes = file_bytes (input_name)) >= 0) {
      if (close_input < 2) assert (actual_input_bytes == read_bytes);
      else msg ("actual input file size %ld bytes", actual_input_bytes);
    }
    if (output_name &&
	strcmp (output_name, "/dev/null") &&
        (actual_output_bytes = file_bytes (output_name)) >= 0) {
      if (close_output < 2) assert (actual_output_bytes == written_bytes);
      else msg ("actual output file size %ld bytes", actual_output_bytes);
      if (close_input > 1 || close_output > 1) {
	if (actual_input_bytes > 0 && actual_output_bytes >= 0)
	  msg ("actual file size written/read bytes ratio %.2f%%",
	    (100.0 * actual_output_bytes) / (double) actual_input_bytes);
      }
    }
  }
  msg ("process time %.2f seconds (including children)", process_time ());
  msg ("maximally %.1f MB allocated", maxbytes / (double)(1<<20));
  msg ("wall clock time %.2f seconds", wall_clock_time ());
}

/*------------------------------------------------------------------------*/

static const char * fmtstr () {
  if (ascii_format) return "ASCII output format";
  else if (binary_format) return "binary output format";
  else return "variable-byte output format";
}

static int isfile (const char * path) {
  struct stat buf;
  if (stat (path, &buf)) return 0;
  return S_ISREG (buf.st_mode);
}

static int hasuffix (const char * str, const char * suffix) {
  int l = strlen (str), k = strlen (suffix);
  if (l < k) return 0;
  return !strcmp (str + l - k, suffix);
}

static char * prepare_pipe_cmd (const char * fmt, 
                                const char * arg, 
				int * bytesptr) {
  int bytes = strlen (fmt) + strlen (arg) + 1;
  char * res;
  res = new (bytes, "pipe command string");
  sprintf (res, fmt, arg);
  *bytesptr = bytes;
  return res;
}

/*------------------------------------------------------------------------*/

static void reset () {
  Cls * p, * n; 
  for (p = first; p; p = n) {
    n = p->next;
    del (p->lits, p->size * sizeof *p->lits);
    del (p, sizeof *p);
  }
  del (table, sztable * sizeof *table);
  del (lits, szlits * sizeof *lits);
  del (order, szorder * sizeof *order);
  del (occs, szorder * sizeof *occs);
  del (map, szorder * sizeof *map);
  assert (!allocated);
}

/*------------------------------------------------------------------------*/

int main (int argc, char ** argv) {
  char * cmd;
  int i;

  start_time = current_time ();

  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h")) {
      fputs (usage, stderr);
      exit (1);
    } else if (!strcmp ("-v", argv[i]))
      verbose_level = 1;
    else if (!strcmp ("-a", argv[i]) || !strcmp ("--ascii", argv[i]))
      ascii_format = 1;
    else if (!strcmp ("-b", argv[i]) || !strcmp ("--binary", argv[i]))
      binary_format = 1;
    else if (!strcmp ("--variable", argv[i]) ||
             !strcmp ("--variable-byte", argv[i]))
      variable_format = 1;
    else if (!strcmp ("--sort", argv[i])) sort_format = 1;
    else if (!strcmp ("--diff", argv[i])) diff_format = 1;
    else if (!strcmp ("--no-sort", argv[i])) no_sort_format = 1;
    else if (!strcmp ("--no-diff", argv[i])) no_diff_format = 1;
    else if (!strcmp ("-r", argv[i]) ||
             !strcmp ("--release", argv[i])) release_mode =1;
    else if (!strcmp ("-o", argv[i]) ||
             !strcmp ("--order", argv[i])) order_mode =1;
    else if (!strcmp ("--big-endian", argv[i])) big_endian_mode = 1;
    else if (argv[i][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[i]);
    else if (output_name)
      die ("unexpected third argument '%s' (try '-h')", argv[i]);
    else if (input_name) output_name = argv[i];
    else input_name = argv[i];
  }

  /*----------------------------------------------------------------------
   * First check and set output format.
   */
  if (ascii_format) {
    if (binary_format) die ("can not use ASCII and binary format");
    if (variable_format) die ("can not use ASCII and variable-bytes format");
    msg ("using ASCII format as specified");
  } else if (binary_format) {
    if (variable_format) die ("can not use binary and variable-byte format");
    msg ("using binary format as specified");
  } else if (variable_format) {
    msg ("using variable-byte format as specified");
  } else if (!output_name && isatty (1)) {
    msg ("using ASCII format (standard output is a terminal)");
    ascii_format = 1;
  } else if (output_name) {
    msg ("using variable-byte format (output file specified)");
    variable_format = 1;
  } else {
    assert (!isatty (1));
    msg ("using variable-byte format (standard output is not a terminal)");
    variable_format = 1;
  }
  assert (ascii_format + binary_format + variable_format == 1);

  /*----------------------------------------------------------------------
   * Do not write binary data to terminal.
   */
  if (!output_name && isatty (1) && !ascii_format)
    die ("can only use ASCII format if output connected to terminal");
  assert (ascii_format + binary_format + variable_format == 1);

  /*----------------------------------------------------------------------
   * Now care about little / big endian.
   */
  if (!binary_format) {
    if (big_endian_mode)
      die ("can only use big-endian with binary format");
  }
  if (big_endian_mode) {
    msg ("using big-endian output mode");
    big_endian_mode = 1;
  } else if (binary_format) {
    msg ("using default little-endian output mode");
    assert (!big_endian_mode);
  }

  /*----------------------------------------------------------------------
   * Handle diff and sort mode.
   */

  /* (2 out of 4) = 6 combinations of options
   */
  if (sort_format && diff_format)
    msg ("ignoring '--sort' because of '--diff'");
  if (sort_format && no_sort_format)
    die ("can not combine '--sort' and '--no-sort'");
  if (sort_format && no_diff_format)
    msg ("ignoring '--no-diff'"), no_diff_format = 0;
  if (diff_format && no_sort_format)
    die ("can not combine '--diff' and '--no-sort'");
  if (diff_format && no_diff_format)
    die ("can not combine '--diff' and '--no-diff'");
  if (no_sort_format && no_diff_format)
    msg ("ignoring '--no-diff' because of '--no-sort'"), no_diff_format = 0;

  if (ascii_format) { /*--------------------------------------------------*/

    if (sort_format)
      msg ("sorting literals in %s ('--sort' specified)", fmtstr ());
    if (diff_format)
      die ("can not use '--diff' in %s", fmtstr ());
    if (no_sort_format)
      msg ("ignoring default '--no-sort' in %s", fmtstr ()),
      no_sort_format = 0;
    if (no_diff_format)
      wrn ("'--no-diff' does not make sense in %s", fmtstr ()),
      no_diff_format = 0;
    assert (sort_format ==
      sort_format + diff_format + no_sort_format + no_diff_format);

  } else if (binary_format) { /*--------------------------------------------*/

    if (diff_format && !sort_format)
      msg ("forcing implicit sorting because of '--diff' in %s",
        fmtstr ()),
      sort_format = 1;
    if (diff_format)
      assert (sort_format),
      msg ("sorting literals and printing differences in %s ('-diff')", 
	   fmtstr ());
    else if (sort_format)
      msg ("sorting and printing literals in %s ('--sort')", fmtstr ());
    else
      msg ("printing literals without sorting in %s (default)", fmtstr ());
    if (no_sort_format)
      msg ("ignoring default '--no-sort' in %s", fmtstr ()),
      no_sort_format = 0;
    if (no_diff_format)
      msg ("ignoring default '--no-diff' in %s", fmtstr ()),
      no_diff_format = 0;
    assert (sort_format + diff_format ==
      sort_format + diff_format + no_sort_format + no_diff_format);
  } else {

    assert (variable_format); /*--------------------------------------------*/

    if (diff_format)
      wrn ("ignoring default '--diff' in %s", fmtstr ()),
      sort_format = 1;
    else if (sort_format)
      wrn ("ignoring '--sort' implied by default '--diff' in %s", fmtstr ()),
      assert (!no_diff_format),
      diff_format = 1;
    else if (no_diff_format)
      assert (!no_sort_format),
      sort_format = 1;
    else if (!no_sort_format)
      diff_format = sort_format = 1;
    else
      assert (!diff_format), assert (!sort_format);
    if (diff_format)
      msg ("sorting literals and printing differences in %s (default)",
        fmtstr ());
    else if (sort_format)
      msg ("sorting and printing literals in %s ('--no-diff')", fmtstr ());
    else
      msg ("printing literals without sorting in %s ('--no-sort')", fmtstr ());
  }

  /*--------------------------------------------------------------------*/
  if (release_mode && ascii_format)
    die ("can not use release for ASCII output format");
  else if (release_mode)
    msg ("using release output format as specified");
  else
    msg ("not using release output format");

  /*--------------------------------------------------------------------*/
  if (order_mode && ascii_format)
    die ("can not order and reencode literals for ASCII output format");
  else if (order_mode)
    msg ("ordering and rencoding literals by number of occurrences");
  else
    msg ("keeping literals in original order");

  /*--------------------------------------------------------------------*/
  /* Determine input file.
   */
  if (input_name) {
    int bytes;
    if (!isfile (input_name)) die ("input '%s' is not a file", input_name);
    if (hasuffix (input_name, ".gz"))
      cmd = prepare_pipe_cmd ("gunzip -c %s", input_name, &bytes);
    else if (hasuffix (input_name, ".bz2"))
      cmd = prepare_pipe_cmd ("bzcat %s", input_name, &bytes);
    else if (hasuffix (input_name, ".7z"))
      cmd = prepare_pipe_cmd (
              "7z x -so %s 2>/dev/null",
	      input_name, &bytes);
    else
      cmd = 0;
    if (cmd) {
      if (!(input_file = popen (cmd, "r")))
	die ("failed to open pipe '%s' for reading", cmd);
      close_input = 2;
      msg ("reading from pipe '%s'", cmd);
      del (cmd, bytes);
    } else {
      if (!(input_file = fopen (input_name, "r")))
	die ("failed to open file '%s' for reading", input_name);
      close_input = 1;
      msg ("reading from file '%s'", input_name);
    } 
  } else {
    assert (!close_input);
    input_name = "<stdin>";
    input_file = stdin;
    msg ("reading from standard input");
  }

  /*--------------------------------------------------------------------*/
  /* Determine output file.
   */
  if (output_name) {
    int bytes;
    if (hasuffix (output_name, ".gz"))
      cmd = prepare_pipe_cmd ("gzip -c > %s", output_name, &bytes);
    else if (hasuffix (output_name, ".bz2"))
      cmd = prepare_pipe_cmd ("bzip2 -c > %s", output_name, &bytes);
    else if (hasuffix (output_name, ".7z"))
      cmd = prepare_pipe_cmd (
              "7z a -an -txz -si -so > %s 2>/dev/null",
	      output_name, &bytes);
    else
      cmd = 0;
    if (cmd) {
      if (!(output_file = popen (cmd, "w")))
	die ("failed to open pipe '%s' for writing", cmd);
      close_output = 2;
      msg ("writing to pipe '%s'", cmd);
      del (cmd, bytes);
    } else {
      if (!(output_file = fopen (output_name, "w")))
	die ("failed to open file '%s' for writing", output_name);
      close_output = 1;
      msg ("writing to file '%s'", output_name);
    }
  } else if (isatty (1) && !ascii_format)
    die ("can not use binary format if output connected to terminal");
  else {
    assert (!close_output);
    output_name = "<stdout>";
    output_file = stdout;
    msg ("writing to standard output");
  }

  /*--------------------------------------------------------------------*/

  transcode ();

  /*--------------------------------------------------------------------*/

  if (close_input == 1) fclose (input_file);
  if (close_input == 2) pclose (input_file);
  if (close_output == 1) fclose (output_file);
  if (close_output == 2) pclose (output_file);
  reset ();
  if (verbose_level) stats ();

  return 0;
}
