/* Copyright (C) 2015, Armin Biere, JKU Linz. */

/* Simple RATZ encoder reading DRAT from stdin and writing DRVT to stdout.
 * This code should only illustrate the logic behind the variable-byte DRVT
 * format and grossly lacks error checking.  Note that half of the code
 * is about parsing the DIMACS like DRAT format.
 */

#include <stdio.h>
#include <stdlib.h>

static int * lits, nlits, szlits;

static void parse_error (int id) {
  fprintf (stderr, "*** sratz: parse error %d\n", id);
  exit (1);
}

static void push_lit (int lit) {
  if (nlits == szlits) {
    if (szlits == (1<<30)) parse_error (1);
    szlits = szlits ? 2*szlits : 1;
    lits = realloc (lits, sizeof *lits * szlits);
  }
  lits[nlits++] = lit;
}

static int cmp (const void * p, const void * q) {
  int a = *(int *) p, b = *(int *) q;
  int res = abs (a) - abs (b);
  if (res) return res;
  return (b > 0) - (a > 0);
}

static unsigned pivot_pos () {
  unsigned pos = 0;
  int i, j, pivot;
  if (!nlits) return 0;
  pivot = lits[0];
  qsort (lits, nlits, sizeof *lits, cmp);
  for (j = i = 1; i < nlits; i++)
    if (lits[j-1] != lits[i]) lits[j++] = lits[i];
  nlits = j;
  while (lits[pos] != pivot) pos++;
  return pos;
}

static void write_unsigned (unsigned w) {
  while (w > 127)
    putc (((w&127)|128), stdout), w >>= 7;
  putc (w, stdout);
}

static void encode (int prev, int next) {
  write_unsigned (2*(abs (next) - abs (prev)) + (next < 0));
}

static void write_lits (int del) {
  int i, prev = 0, lit;
  write_unsigned ((pivot_pos () << 2) | !del);
  for (i = 0; i < nlits; i++, prev = lit)
    encode (prev, lit = lits[i]);
  putc (0, stdout);
}

int main () {
  int ch, del = 0;
  for (;;) {
    int lit;
    ch = getc (stdin);
    if (ch == 'd') { del = 1; continue; }
    if (ch == ' ' || ch == '\n') continue;
    if (ch == EOF) { if (del || nlits) parse_error (2); break; }
    if (ch == 'c' || ch == 'p' || ch == 'v' || ch == 's') {
      while ((ch = getc (stdin)) != '\n')
        if (ch == EOF) parse_error (3);
      continue;
    }
    ungetc (ch, stdin);
    if (fscanf (stdin, "%d", &lit) != 1) parse_error (4);
    else if (lit) push_lit (lit);
    else write_lits (del), nlits = del = 0;
  }
  free (lits);
  return 0;
}
