/* Copyright (C) 2015, Armin Biere, JKU Linz. */

static const char * usage =
"usage: ztar [ <option> ... ] [ <input> [ <output> ] ]\n"
"\n"
"where <option> is one of the following:\n"
"\n"
"-h | --help   print this command line option summary\n"
"-v            enable printing of verbose messages\n"
"\n"
"The <input> is supposed to be a binary or variable DRAT proof file and is\n"
"written in ASCII format to the output file.  If <output> is not specified\n"
"the standard output is used and similarly if <input> is not specified then\n"
"the standard input is used.  If a file is specified and has a name with\n"
"'.gz', '.bz2', or '.7z' suffix, then input or output are decompressed or\n"
"compressed using helper programs through a pipe.\n"
;

/*------------------------------------------------------------------------*/

#include "typez.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/

typedef struct Cls { int size, * lits; } Cls;

/*------------------------------------------------------------------------*/

static char * input_name, * output_name;
static FILE * input_file, * output_file;
static int close_input, close_output;

static int binary_format, variable_format;
static int diff_mode, big_endian_mode, release_mode;
static int verbose_level;

static int * lits, nlits, szlits, pivot_pos;

static Cls ** clauses;
static unsigned nclauses, szclauses;
static int active, maxcls;

static int saved_char, saved_char_valid;

static long read_bytes, written_bytes;
static long added_clauses, deleted_clauses, released_clauses;
static long added_lits, deleted_lits, released_lits;

static int sort_format = 1;

static double start_time;
static long allocated, maxbytes;

/*------------------------------------------------------------------------*/

static void die (const char * msg, ...) {
  va_list ap;
  fputs ("*** ztar: error: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static void msg (const char * msg, ...) {
  va_list ap;
  if (!verbose_level) return;
  fputs ("[ztar] ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
}

static void perr (const char * msg, ...) {
  va_list ap;
  fprintf (stderr,
    "*** ztar: parse error after reading %ld bytes of '%s': ",
    read_bytes, input_name);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

/*------------------------------------------------------------------------*/

static void incmem (size_t bytes) {
  allocated += bytes;
  if (maxbytes < allocated) maxbytes = allocated;
}

static void decmem (size_t bytes) {
  assert (allocated >= bytes);
  allocated -= bytes;
}

static void * new (size_t bytes, const char * what) {
  void * res = malloc (bytes);
  if (bytes && !res) die ("failed to allocate %s", what);
  incmem (bytes);
  return res;
}

static void del (void * ptr, size_t bytes) {
  decmem (bytes);
  free (ptr);
}

static void * rsz (void * ptr, 
                   size_t oldbytes, size_t newbytes,
		   const char * what) {
  void * res;
  decmem (oldbytes);
  res = realloc (ptr, newbytes);
  if (newbytes && !res) die ("failed to reallocate %s", what);
  incmem (newbytes);
  return res;
}

/*------------------------------------------------------------------------*/

static void push_lit (int lit) {
  if (nlits >= szlits) {
    size_t oldbytes, newbytes;
    if (szlits == (1<<30))
      die ("clause too large (> 2^30 literals)");
    oldbytes = szlits * sizeof *lits;
    szlits = szlits ? 2*szlits : 1;
    newbytes = szlits * sizeof *lits;
    lits = rsz (lits, oldbytes, newbytes, "literal stack");
  }
  lits[nlits++] = lit;
}

/*------------------------------------------------------------------------*/

static void add_clause () {
  size_t bytes;
  Cls * c;
  if (!release_mode) return;
  if (nclauses >= szclauses) {
    size_t oldbytes;
    if (szclauses == (1<<30)) die ("too many clauses (> 2^30)");
    oldbytes = szclauses * sizeof *clauses;
    szclauses = szclauses ? 2*szclauses : 1;
    bytes = szclauses * sizeof *clauses;
    clauses =  rsz (clauses, oldbytes, bytes, "clause stack");
  }
  c = new (sizeof *c, "clause");
  c->size = nlits;
  bytes = nlits * sizeof *c->lits;
  c->lits = new (bytes, "clause lits");
  memcpy (c->lits, lits, bytes);
  clauses[nclauses++] = c;
  if (++active > maxcls) maxcls = active;
}

static void delete_clause (Cls * c) {
  del (c->lits, c->size * sizeof *c->lits);
  del (c, sizeof *c);
}

/*------------------------------------------------------------------------*/

static int read_char () {
  int res;
  if (saved_char_valid) res = saved_char, saved_char_valid = 0;
  else res = getc_unlocked (input_file);
  if (res != EOF) read_bytes++;
  return res;
}

static void save_char (int ch) {
  assert (!saved_char_valid);
  saved_char = ch;
  saved_char_valid = 1;
  if (ch == EOF) return;
  assert (read_bytes > 0);
  read_bytes--;
}

static void write_char (unsigned char ch) {
  if (putc_unlocked ((int) ch, output_file) == EOF)
    die ("failed to write byte %ld", written_bytes + 1);
  written_bytes++;
}

static void write_str (const char * str) {
  const char *p;
  for (p = str; *p; p++)
    write_char (*p);
}

static void write_lit (int lit) {
  char buf[20];
  sprintf (buf, "%d", lit);
  write_str (buf);
}

/*------------------------------------------------------------------------*/

static unsigned binary_read_big_endian_word () {
  unsigned res = 0;
  int i, ch;
  for (i = 3; i >= 0; i--) {
    if ((ch = read_char ()) == EOF)
perr ("unexpected end-of-file reading big-endian byte %d of word", i);
    res |= ((unsigned char) (ch & 0xff)) << (i*8);
  }
  return res;
}

static unsigned binary_read_little_endian_word () {
  unsigned res = 0;
  int i, ch;
  for (i = 0; i < 4; i++) {
    if ((ch = read_char ()) == EOF)
perr ("unexpected end-of-file reading little-endian byte %d of word", i);
    res |= ((unsigned char) (ch & 0xff)) << (i*8);
  }
  return res;
}

static unsigned binary_read_word () {
  assert (binary_format);
  if (big_endian_mode) return binary_read_big_endian_word ();
  else return binary_read_little_endian_word ();
}

/*------------------------------------------------------------------------*/

static unsigned variable_read_word () {
  unsigned char ch;
  unsigned res = 0;
  int i;
  assert (variable_format);
  i = 0;
  for (;;) {
    if ((ch = read_char ()) == EOF)
      perr ("unexpected end-of-file reading variable-byte %d", i);
    if (i == 28 && ch > 15)
      perr ("highest 4 bits non-zero in variable-byte 5");
    assert (i < 32);
    res |= (ch & 127) << i;
    if (ch < 128) break;
    if (i == 28) perr ("more than 5 bytes in variable-byte encoding");
    i += 7;
  }
  return res;
}

/*------------------------------------------------------------------------*/

static unsigned read_word () {
  if (binary_format) return binary_read_word ();
  else return variable_read_word ();
}

static int decode_lit (int prev, unsigned w) {
  int res = w/2;
  assert (res >= 0);
  if (prev && diff_mode) {
    int absprev = abs (prev);
    if (INT_MAX - res < absprev)
      perr ("previous literal %d + delta %d > INT_MAX", prev, res);
    res += absprev;
  }
  if (w & 1) res = -res;
  if (sort_format && prev && abs (prev) >= abs (res)) sort_format = 0;
  return res;
}

static int decode_type_and_pivot_pos () {
  unsigned w = read_word ();
  int type = w & RATZ_TYPE_MASK;
  pivot_pos = w >> RATZ_TYPE_BITS;
  return type;
}

/*------------------------------------------------------------------------*/

static int read_lits () {
  int type, lit = 0;
  const int * p, * eol;
  unsigned w;
  assert (!nlits);
  type = decode_type_and_pivot_pos ();
  if (type == RATZ_REL_TYPE) {
    unsigned did;
    Cls * c;
    w = read_word ();
    if (!w) perr ("invalid zero release reference");
    if (w > nclauses)
      perr ("invalid too large release reference %u", w);
    did = nclauses - w;
    c = clauses[did];
    if (!c) perr ("non existing release reference %u", w);
    eol = c->lits + c->size;
    for (p = c->lits; p < eol; p++) push_lit (*p);
    clauses[did] = 0;
    assert (active > 0);
    active--;
  } else {
    while ((w = read_word ()))
      push_lit (lit = decode_lit (lit, w));
  }
  if (!nlits && pivot_pos)
    perr ("unexpected pivot position %d for empty clause");
  if (nlits && pivot_pos >= nlits)
    perr ("pivot position %d >= clause length %d", pivot_pos, nlits);
  return type;
}

static void write_lits (int type) {
  int i;
  if (type == RATZ_DEL_TYPE || type == RATZ_REL_TYPE) write_str ("d ");
  if (nlits) write_lit (lits[pivot_pos]), write_char (' ');
  for (i = 0; i < nlits; i++)
    if (i != pivot_pos)
      write_lit (lits[i]), write_char (' ');
  write_lit (0);
  write_char ('\n');
}

static int transcode_lits () {
  int type, ch = read_char ();
  if (ch == EOF) return 0;
  save_char (ch);
  type = read_lits ();
  write_lits (type);
  if (type == RATZ_REL_TYPE) released_clauses++, released_lits += nlits;
  else if (type == RATZ_DEL_TYPE) deleted_clauses++, deleted_lits += nlits;
  else add_clause (), added_clauses++, added_lits += nlits;
  nlits = 0;
  return 1;
}

/*------------------------------------------------------------------------*/

static void transcode_header () {
  int ch = read_char ();
  if (ch == EOF || !(ch & RATZ_HEADER_BIT)) {
    msg ("no header: variable-byte differences, no release");
    variable_format = 1;
    diff_mode = 1;
    save_char (ch);
  } else if (ch & ~RATZ_HEADER_MASK)
    perr ("illegal bits in header byte code 0x%02x", ch);
  else {
    if (ch & RATZ_BINARY_HEADER_BIT) {
      binary_format = 1;
      msg ("header specifies binary input format");
    } else {
      variable_format = 1;
      msg ("header specifies variable input format");
    }
    if (binary_format) {
      if (ch & RATZ_BIG_ENDIAN_HEADER_BIT) {
	big_endian_mode = 1;
	msg ("header specifies big-endian mode");
      } else {
	assert (!big_endian_mode);
	msg ("header specifies little-endian mode");
      }
    } else if (ch & RATZ_BIG_ENDIAN_HEADER_BIT)
      perr ("header with illegal big-endian bit in variable-byte format");
    if (ch & RATZ_NO_DIFFERENCES_HEADER_BIT) {
      assert (!diff_mode);
      msg ("header specifies literal mode");
    } else {
      diff_mode = 1;
      msg ("header specifies difference mode");
    }
    if (ch & RATZ_RELEASE_HEADER_BIT) {
      release_mode = 1;
      msg ("header specifies release mode");
    } else {
      assert (!release_mode);
      msg ("header specifies no release mode");
    }
  }
}

static void transcode () {
  transcode_header ();
  while (transcode_lits ())
    ;
}

/*------------------------------------------------------------------------*/

static long file_bytes (const char * name) {
  struct stat buf;
  if (stat (name, &buf)) return -1;
  return (long) buf.st_size;
}

static double current_time (void) {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0)) res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static double wall_clock_time () { 
  return current_time () - start_time; 
}

static double process_time (void) {
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  if (getrusage (RUSAGE_CHILDREN, &u)) return 0;
  res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

static void stats () {
  long actual_input_bytes, actual_output_bytes;
  long sum = deleted_clauses + released_clauses;
  msg ("added %ld clauses, deleted %ld clauses",
    added_clauses, deleted_clauses);
  if (release_mode) {
    msg ("released %ld clauses, %.0f%% of added, %.0f%% of deleted",
    released_clauses,
    added_clauses ? 100.0 * released_clauses/(double)sum : 0,
    sum ? 100.0 * released_clauses/(double)sum : 0);
    msg ("maximum number %d of active clauses %.0f%%",
      maxcls, added_clauses ? 100.0 * maxcls / (double) added_clauses : 0);
  }
  msg ("added %ld literals, %.1f per added clause",
    added_lits, added_clauses ? added_lits / (double) added_clauses : 0);
  msg ("deleted %ld literals, %.1f per deleted clause",
    deleted_lits,
    deleted_clauses ? deleted_lits / (double) deleted_clauses : 0);
  if (release_mode)
    msg ("released %ld literals, %.1f per released clause",
      released_lits,
      released_clauses ? released_lits / (double) released_clauses : 0);
  msg ("read %ld bytes, written %ld bytes",
    read_bytes, written_bytes);
  if (read_bytes)
    msg ("written/read bytes ratio %.2f%%",
      (100.0 * written_bytes) / (double) read_bytes);
  if (input_name) {
    assert (close_input);
    if ((actual_input_bytes = file_bytes (input_name)) >= 0) {
      if (close_input < 2) assert (actual_input_bytes == read_bytes);
      else msg ("actual input file size %ld bytes", actual_input_bytes);
    }
    if (output_name &&
        (actual_output_bytes = file_bytes (output_name)) >= 0) {
      if (close_output < 2) assert (actual_output_bytes == written_bytes);
      else msg ("actual output file size %ld bytes", actual_output_bytes);
      if (close_input > 1 || close_output > 1) {
	if (actual_input_bytes > 0 && actual_output_bytes >= 0)
	  msg ("actual file size written/read bytes ratio %.2f",
	    actual_output_bytes / (double) actual_input_bytes);
      }
    }
  }
  msg ("process time %.2f seconds (including children)", process_time ());
  msg ("maximally %.1f MB allocated", maxbytes / (double)(1<<20));
  msg ("wall clock time %.2f seconds", wall_clock_time ());
}

/*------------------------------------------------------------------------*/

static int isfile (const char * path) {
  struct stat buf;
  if (stat (path, &buf)) return 0;
  return S_ISREG (buf.st_mode);
}

static int hasuffix (const char * str, const char * suffix) {
  int l = strlen (str), k = strlen (suffix);
  if (l < k) return 0;
  return !strcmp (str + l - k, suffix);
}

static char * prepare_pipe_cmd (const char * fmt, 
                                const char * arg, 
				int * bytesptr) {
  int bytes = strlen (fmt) + strlen (arg) + 1;
  char * res;
  res = new (bytes, "pipe command string");
  sprintf (res, fmt, arg);
  *bytesptr = bytes;
  return res;
}

/*------------------------------------------------------------------------*/

static void reset () {
  unsigned i;
  for (i = 0; i < nclauses; i++)
    if (clauses[i]) delete_clause (clauses[i]);
  del (clauses, szclauses * sizeof *clauses);
  del (lits, szlits * sizeof *lits);
}

/*------------------------------------------------------------------------*/

int main (int argc, char ** argv) {
  char * cmd;
  int i;

  start_time = current_time ();

  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h")) {
      fputs (usage, stderr);
      exit (1);
    } else if (!strcmp ("-v", argv[i]))
      verbose_level = 1;
    else if (argv[i][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[i]);
    else if (output_name)
      die ("unexpected third argument '%s' (try '-h')", argv[i]);
    else if (input_name) output_name = argv[i];
    else input_name = argv[i];
  }

  /*----------------------------------------------------------------------*/
  /* First setup input file.
   */
  if (input_name) {
    int bytes;
    if (!isfile (input_name)) die ("input '%s' is not a file", input_name);
    if (hasuffix (input_name, ".gz"))
      cmd = prepare_pipe_cmd ("gunzip -c %s", input_name, &bytes);
    else if (hasuffix (input_name, ".bz2"))
      cmd = prepare_pipe_cmd ("bzcat %s", input_name, &bytes);
    else if (hasuffix (input_name, ".7z"))
      cmd = prepare_pipe_cmd (
              "7z x -so %s 2>/dev/null",
	      input_name, &bytes);
    else
      cmd = 0;
    if (cmd) {
      if (!(input_file = popen (cmd, "r")))
	die ("failed to open pipe '%s' for reading", cmd);
      close_input = 2;
      msg ("reading from pipe '%s'", cmd);
      del (cmd, bytes);
    } else {
      if (!(input_file = fopen (input_name, "r")))
	die ("failed to open file '%s' for reading", input_name);
      close_input = 1;
      msg ("reading from file '%s'", input_name);
    } 
  } else {
    assert (!close_input);
    input_name = "<stdin>";
    input_file = stdin;
    msg ("reading from standard input");
  }

  /*----------------------------------------------------------------------*/
  /* Then setup output file.
   */
  if (output_name) {
    int bytes;
    if (hasuffix (output_name, ".gz"))
      cmd = prepare_pipe_cmd ("gzip -c > %s", output_name, &bytes);
    else if (hasuffix (output_name, ".bz2"))
      cmd = prepare_pipe_cmd ("bzip2 -c > %s", output_name, &bytes);
    else if (hasuffix (output_name, ".7z"))
      cmd = prepare_pipe_cmd (
              "7z a -an -txz -si -so > %s 2>/dev/null",
	      output_name, &bytes);
    else
      cmd = 0;
    if (cmd) {
      if (!(output_file = popen (cmd, "w")))
	die ("failed to open pipe '%s' for writing", cmd);
      close_output = 2;
      msg ("writing to pipe '%s'", cmd);
      del (cmd, bytes);
    } else {
      if (!(output_file = fopen (output_name, "w")))
	die ("failed to open file '%s' for writing", output_name);
      close_output = 1;
      msg ("writing to file '%s'", output_name);
    }
  } else {
    assert (!close_output);
    output_name = "<stdout>";
    output_file = stdout;
    msg ("writing to standard output");
  }

  /*----------------------------------------------------------------------*/

  transcode ();

  /*----------------------------------------------------------------------*/

  if (close_input == 1) fclose (input_file);
  if (close_input == 2) pclose (input_file);
  if (close_output == 1) fclose (output_file);
  if (close_output == 2) pclose (output_file);

  if (sort_format) msg ("input was probably sorted");
  else msg ("input was definitely not sorted");
  reset ();
  if (verbose_level) stats ();

  return 0;
}
